from .. import __version__ as nbs_version
from .gen import generate_gallery_rst, DEFAULT_GALLERY_CONF

def setup(app):
    """Setup sphinx-gallery sphinx extension"""
    app.add_config_value('nbsite_gallery_conf', DEFAULT_GALLERY_CONF, 'html')

    app.connect('builder-inited', generate_gallery_rst)
    metadata = {'parallel_read_safe': True,
                'version': nbs_version}
    return metadata



#!/usr/bin/env python

import sys
import os

from nbsite.cmd import generate_rst

# if only someone had made a way to handle parameters
org = sys.argv[1]
project = sys.argv[2]
examples_path = os.path.abspath(sys.argv[3])
doc_path = os.path.abspath(sys.argv[4])
offset = 0
overwrite = bool(1)
if len(sys.argv) > 5:
    offset = int(sys.argv[5])
if len(sys.argv) > 6:
    overwrite = int(sys.argv[6])
    
generate_rst(
    project,
    examples=examples_path,
    doc=doc_path,
    org = org,
    repo=project,
    offset=offset,
    overwrite=bool(overwrite))



import argparse
import inspect

from .cmd import build, generate_rst, init


def _add_common_args(parser,*names):
    common = {
        '--project-root': dict(type=str,help='defaults to current working directory',default=''),
        '--examples': dict(type=str,help='if relative, should be relative to project-root',default='examples'),
        '--doc': dict(type=str,help='if relative, should be releative to project-root',default='doc'),
        '--overwrite': dict(action='store_true', help='whether to overwrite any files [DANGEROUS]')
        }
    for name in names:
        parser.add_argument(name,**common[name])

def _set_defaults(parser,fn):
    parser.set_defaults(func=lambda args: fn( **{k: getattr(args,k) for k in vars(args) if k!='func'} ))

def main(args=None):
    parser = argparse.ArgumentParser(description="nbsite commands")
    subparsers = parser.add_subparsers(title='available commands')

    init_parser = subparsers.add_parser("init", help=inspect.getdoc(init))
    _add_common_args(init_parser,'--project-root','--doc')
    init_parser.add_argument('--theme', type=str, help='sphinx theme to use in template', choices=['holoviz', ''], default='')
    _set_defaults(init_parser,init)

    generaterst_parser = subparsers.add_parser("generate-rst", help=inspect.getdoc(generate_rst))
    _add_common_args(generaterst_parser,'--project-root','--doc','--examples', '--overwrite')
    generaterst_parser.add_argument('--project-name', type=str, help='name of project')
    generaterst_parser.add_argument('--host',type=str,help='host to use when generating notebook links',default='GitHub')
    generaterst_parser.add_argument('--org',type=str,help='github organization',default='')
    generaterst_parser.add_argument('--repo',type=str,help='name of repo',default='')
    generaterst_parser.add_argument('--branch',type=str,help='branch to point to in notebook links',default='main')
    generaterst_parser.add_argument('--offset',type=int,help='number of cells to delete from top of notebooks',default=0)
    generaterst_parser.add_argument('--nblink',type=str,help='where to place notebook links',choices=['bottom', 'top', 'both', 'none'], default='bottom')
    generaterst_parser.add_argument('--binder',type=str,help='where to place binder link',choices=['bottom', 'top', 'both', 'none'], default='none')
    generaterst_parser.add_argument('--skip',type=str,help='notebooks to skip running; comma separated case insensitive re to match',default='')
    generaterst_parser.add_argument('--keep-numbers',action='store_true',help='whether to keep the leading numbers of notebook URLs and titles')
    generaterst_parser.add_argument('--disable-interactivity-warning',action='store_true',help='whether to disable interactivity warnings')
    _set_defaults(generaterst_parser,generate_rst)

    build_parser = subparsers.add_parser("build", help=inspect.getdoc(build))
    build_parser.add_argument('--what',type=str,help='type of output to generate',default='html')

    build_parser.add_argument('--project-name', type=str, help='name of project', default='')
    build_parser.add_argument('--org',type=str,help='github organization',default='')
    build_parser.add_argument('--host',type=str,help='host to use when generating notebook links',default='GitHub')
    build_parser.add_argument('--repo',type=str,help='name of repo',default='')
    build_parser.add_argument('--branch',type=str,help='branch to point to in notebook links',default='main')
    build_parser.add_argument('--binder',type=str,help='where to place binder link',choices=['bottom', 'top', 'both', 'none'], default='none')

    build_parser.add_argument('--output',type=str,help='where to place output',default="builtdocs")
    _add_common_args(build_parser,'--project-root','--doc','--examples', '--overwrite')
    build_parser.add_argument('--examples-assets',type=str,default="assets",
                              help='dir in which assets for examples are located - if relative, should be releative to project-root')
    build_parser.add_argument('--clean-dry-run',action='store_true',help='whether to not actually delete files from output (useful for uploading)')
    build_parser.add_argument('--inspect-links',action='store_true',help='whether to not to print all links')
    _set_defaults(build_parser,build)

    args = parser.parse_args()
    return args.func(args) if hasattr(args,'func') else parser.error("must supply command to run")

if __name__ == "__main__":
    main()



import importlib
import io
import json
import re
import warnings
from collections import defaultdict
from html import escape
from multiprocessing import Pipe, Process
from pathlib import Path
from typing import Any, Dict, List, Tuple

from bokeh.document import Document
from bokeh.embed.util import standalone_docs_json_and_render_items
from bokeh.model import Model
from docutils import nodes
from docutils.parsers.rst import Directive, roles
from jinja2.environment import Environment
from jinja2.loaders import FileSystemLoader
from packaging.version import Version
from panel.config import panel_extension
from panel.io.convert import BOKEH_VERSION
from panel.io.mime_render import exec_with_return, format_mime
from panel.io.resources import CDN_DIST, set_resource_mode
from panel.pane import panel as as_panel
from panel.util import is_holoviews
from panel.viewable import Viewable, Viewer
from sphinx.application import Sphinx

HERE = Path(__file__).parent

def get_env() -> Environment:
    ''' Get the correct Jinja2 Environment, also for frozen scripts.
    '''
    return Environment(loader=FileSystemLoader(HERE))

_env = get_env()
_env.trim_blocks = True
_env.lstrip_blocks = True

WEB_WORKER_TEMPLATE = _env.get_template('WebWorker.js')
WORKER_HANDLER_TEMPLATE = _env.get_template('WorkerHandler.js')
SERVICE_WORKER_TEMPLATE = _env.get_template('ServiceWorker.js')
SERVICE_HANDLER_TEMPLATE = _env.get_template('ServiceHandler.js')
WEB_MANIFEST_TEMPLATE = _env.get_template('site.webmanifest')

JS_MODULE_EXPORT = """
<script type="module">
  import {name} from "{file}";
  window.{name} = {name}
</script>
"""

bokeh_version = Version(BOKEH_VERSION)
if bokeh_version.is_devrelease or bokeh_version.is_prerelease:
    bk_prefix = 'dev'
else:
    bk_prefix = 'release'

DEFAULT_PYODIDE_CONF = {
    'PYODIDE_URL': 'https://cdn.jsdelivr.net/pyodide/v0.22.1/full/pyodide.js',
    'autodetect_deps': True,
    'enable_pwa': True,
    'requirements': ['panel', 'pandas'],
    'precache': [],
    'scripts': [
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-{BOKEH_VERSION}.min.js',
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-widgets-{BOKEH_VERSION}.min.js',
        f'https://cdn.bokeh.org/bokeh/{bk_prefix}/bokeh-tables-{BOKEH_VERSION}.min.js',
        f'{CDN_DIST}panel.min.js'
    ],
    'extra_css': [],
    'cache_patterns': [
        CDN_DIST,
        'https://cdn.bokeh.org/bokeh/',
        'https://cdn.jsdelivr.net/pyodide/',
        'https://files.pythonhosted.org/packages/',
        'https://pypi.org/pypi/'
    ],
    'setup_code': "",
    'warn_message': "Executing this cell will download Python runtime (typically 40+ MB)."
}

EXTRA_RESOURCES = defaultdict(lambda: {'js': [], 'css': [], 'js_modules': {}})

def extract_extensions(code: str) -> List[str]:
    """
    Extracts JS and CSS dependencies of Panel extensions from code snippets
    containing pn.extension calls.
    """
    ext_args = re.findall(r'pn.extension\((.*)\)', code)
    if not ext_args:
        return [], {}, []
    extensions = re.findall(r"(?<!=)\s*['\"](.*?)['\"]", ext_args[0])
    prev_models = dict(Model.model_class_reverse_map)
    for ext in extensions:
        if ext not in panel_extension._imports:
            continue
        importlib.import_module(panel_extension._imports[ext])
    js, js_modules, css = [], {}, []
    with set_resource_mode('cdn'):
        for name, model in Model.model_class_reverse_map.items():
            if name not in prev_models:
                if hasattr(model, '__javascript__'):
                    js += model.__javascript__
                if hasattr(model, '__css__'):
                    css += model.__css__
                if hasattr(model, '__javascript_module_exports__'):
                    js_modules.update(
                        dict(zip(model.__javascript_module_exports__,
                                 model.__javascript_modules__))
                    )
    return js, js_modules, css

def _model_json(model: Model, target: str) -> Tuple[Document, str]:
    """
    Renders a Bokeh Model to JSON representation given a particular
    DOM target and returns the Document and the serialized JSON string.

    Arguments
    ---------
    model: bokeh.model.Model
        The bokeh model to render.
    target: str
        The id of the DOM node to render to.

    Returns
    -------
    document: Document
        The bokeh Document containing the rendered Bokeh Model.
    model_json: str
        The serialized JSON representation of the Bokeh Model.
    """
    doc = Document()
    model.server_doc(doc=doc)
    model = doc.roots[0]
    docs_json, _ = standalone_docs_json_and_render_items(
        [model], suppress_callback_warning=True
    )
    doc_json = list(docs_json.values())[0]

    return doc, json.dumps(dict(
        target_id = target,
        root_id   = model.ref['id'],
        doc       = doc_json,
        version   = BOKEH_VERSION,
    ))


def _option_boolean(arg):
    if not arg or not arg.strip():
        # no argument given, assume used as a flag
        return True
    elif arg.strip().lower() in ("no", "0", "false"):
        return False
    elif arg.strip().lower() in ("yes", "1", "true"):
        return True
    else:
        raise ValueError('"%s" unknown boolean' % arg)

class PyodideDirective(Directive):

    has_content = True

    option_spec = {
        'skip-embed': _option_boolean
    }

    _current_source = None
    _current_context = {}
    _current_count = 0
    _current_process = None
    _conn = None

    @classmethod
    def _execution_process(cls, pipe):
        """
        Process execution loop to run in separate process that is used
        to evaluate code.
        """
        while True:
            msg = pipe.recv()
            if msg['type'] == 'close':
                break
            elif msg['type'] != 'execute':
                continue
            stdout = io.StringIO()
            stderr = io.StringIO()
            code = msg['code']
            js, js_modules, css = extract_extensions(code)
            with set_resource_mode('cdn'):
                try:
                    out = exec_with_return(code, stdout=stdout, stderr=stderr)
                except Exception:
                    out = None
                if isinstance(out, (Model, Viewable, Viewer)) or is_holoviews(out):
                    _, content = _model_json(as_panel(out), msg['target'])
                    mime_type = 'application/bokeh'
                elif out is not None:
                    try:
                        content, mime_type = format_mime(out)
                    except Exception:
                        warnings.warn(f'Could not render {out!r} generated from executed code directive: {code}')
                else:
                    content, mime_type = None, None
            pipe.send((content, mime_type, stdout.getvalue(), stderr.getvalue(), js, js_modules, css))
        pipe.close()

    @classmethod
    def terminate(cls, *args):
        """
        Terminates a running process.
        """
        if not cls._current_process:
            return
        try:
            cls._conn.send({'type': 'close'})
            cls._current_process.join()
        except Exception:
            cls._current_process.terminate()
        finally:
            cls._current_process = None

    @classmethod
    def _launch_process(cls, timeout=5):
        """
        Launches a process to execute code in.
        """
        cls.terminate()
        cls._conn, child_conn = Pipe()
        cls._current_process = Process(target=cls._execution_process, args=(child_conn,))
        cls._current_process.start()

    @classmethod
    def _kill(cls):
        cls._current_process.terminate()
        cls._current_process = None

    def run(self):
        current_source = self.state_machine.get_source()
        if self._current_source != current_source or self._current_process is None:
            PyodideDirective._current_count = 0
            PyodideDirective._current_source = current_source
            self._launch_process()

        classes = 'pyodide'
        if 'class' in self.options:
            classes += f" {self.options['class']}"
        self.options['class'] = [classes]
        self.options['id'] = cellid = f'codecell{self._current_count}-py'
        roles.set_classes(self.options)
        code = '\n'.join(self.content)
        doctree_node = nodes.literal_block(code, code, **self.options)
        doctree_node['language'] = 'python'

        PyodideDirective._current_count += 1
        if self.options.get('skip-embed'):
            return [doctree_node]

        # Send execution request to client and wait for result
        self._conn.send({'type': 'execute', 'target': f'output-{cellid}', 'code': code})
        if self._conn.poll(60):
            try:
                output, mime_type, stdout, stderr, js, js_modules, css = self._conn.recv()
            except Exception:
                self._kill()
                return [doctree_node]
        else:
            self._kill()
            return [doctree_node]
        EXTRA_RESOURCES[current_source]['js'] += js
        EXTRA_RESOURCES[current_source]['js_modules'].update(js_modules)
        EXTRA_RESOURCES[current_source]['css'] += css

        stdout_style = 'style="display: block;"' if stdout else ''
        stdout_html = f'<pre id="stdout-{cellid}" class="pyodide-stdout" {stdout_style}>{escape(stdout)}</pre>'
        stdout_node = nodes.raw(stdout, stdout_html, format='html')

        stderr_style = 'style="display: block;"' if stderr else ''
        stderr_html = f'<pre id="stderr-{cellid}" class="pyodide-stderr" {stderr_style}>{escape(stderr)}</pre>'
        stderr_node = nodes.raw(stderr, stderr_html, format='html')

        rendered_nodes = [doctree_node, stdout_node, stderr_node]
        if output is None:
            return rendered_nodes

        # Ensure we wait for all JS module exports to be initialized
        exports = ' && '.join(f'window.{export}' for export in EXTRA_RESOURCES[current_source]['js_modules'])
        if exports:
            exports = f' && {exports}'

        script = ""
        if mime_type == 'text/plain':
            output = f'<pre>{output}</pre>'
        elif mime_type == 'application/bokeh':
            script = f"""
            <script>
              async function embed_bokeh_{self._current_count} () {{
                if (window.Bokeh && window.Bokeh.Panel{exports}) {{
                  await Bokeh.embed.embed_item({output})
                }} else {{
                   setTimeout(embed_bokeh_{self._current_count}, 200)
                }}
              }};
              embed_bokeh_{self._current_count}()
            </script>
            """
            output = ""
        else:
            output = f'<div class="pyodide-output-wrapper">{output}</div>'
        output_html = f'<div id="output-{cellid}" class="pyodide-output embedded">{output}</div>\n{script}'
        output_node = nodes.raw(output, output_html, format='html')
        return rendered_nodes+[output_node]

def write_worker(app: Sphinx, exc):
    if app.builder.format != 'html' or exc:
        return
    pyodide_conf = app.config.nbsite_pyodide_conf
    builddir = Path(app.builder.outdir)
    staticdir = builddir / '_static'

    # Render Web Worker
    web_worker = WEB_WORKER_TEMPLATE.render({
        'PYODIDE_URL': pyodide_conf['PYODIDE_URL'],
        'env_spec': ', '.join([repr(req) for req in pyodide_conf['requirements']]),
        'setup_code': pyodide_conf['setup_code'],
        'autodetect_deps': pyodide_conf['autodetect_deps']
    })
    with open(staticdir/ 'PyodideWebWorker.js', 'w') as f:
        f.write(web_worker)
    worker_setup = WORKER_HANDLER_TEMPLATE.render(
        scripts=pyodide_conf['scripts']
    )
    with open(staticdir/ 'WorkerHandler.js', 'w') as f:
        f.write(worker_setup)

    if not pyodide_conf['enable_pwa']:
        return

    # Render service worker
    service_worker = SERVICE_WORKER_TEMPLATE.render({
        'project': app.config.project,
        'version': app.config.version,
        'pre_cache': ', '.join([repr(req) for req in pyodide_conf['precache']]),
        'cache_patterns': ', '.join([repr(req) for req in pyodide_conf['cache_patterns']])
    })
    with open(builddir / 'PyodideServiceWorker.js', 'w') as f:
        f.write(service_worker)
    service_handler = SERVICE_HANDLER_TEMPLATE.render()
    with open(staticdir/ 'ServiceHandler.js', 'w') as f:
        f.write(service_handler)

    # Render manifest
    site_manifest = WEB_MANIFEST_TEMPLATE.render({
        'name': app.config.html_title,
    })
    with open(builddir / 'site.webmanifest', 'w') as f:
        f.write(site_manifest)


def init_conf(app: Sphinx) -> None:
    pyodide_conf = dict(DEFAULT_PYODIDE_CONF, **app.config.nbsite_pyodide_conf)
    app.config.nbsite_pyodide_conf = pyodide_conf
    app.config.html_static_path.append(
        str((HERE /'_static' ).absolute())
    )
    if pyodide_conf['enable_pwa']:
        app.add_js_file('ServiceHandler.js')
    for script in pyodide_conf['scripts']:
        app.add_js_file(script)

def html_page_context(
    app: Sphinx,
    pagename: str,
    templatename: str,
    context: Dict[str, Any],
    doctree: nodes.document,
) -> None:
    if not doctree:
        return

    if app.config.nbsite_pyodide_conf['enable_pwa']:
        relpath = '/'.join(['..']*pagename.count('/'))
        if relpath:
            relpath += '/'
        context[
            "metatags"
        ] += f"""
            <link rel="manifest" href="{relpath}site.webmanifest"/>
            """

    # Add additional resources extracted from pn.extension calls
    sourcename = context['sourcename'].replace('.txt', '')
    extra_resources = [
        (r['js'], r['js_modules'], r['css']) for filename, r in EXTRA_RESOURCES.items()
        if filename.endswith(sourcename)
    ]
    if extra_resources:
        extra_js, js_modules, extra_css = extra_resources[0]
    else:
        extra_js, js_modules, extra_css = [], {}, []
    extra_css += app.config.nbsite_pyodide_conf.get('extra_css', [])
    context["script_files"] += extra_js
    context["css_files"] += extra_css

    module_tags = ""
    for export, module in js_modules.items():
        module_tags += JS_MODULE_EXPORT.format(name=export, file=module)
    if module_tags:
        context["body"] = f'{module_tags}{context["body"]}'

    # Remove Scripts and CSS from page if no pyodide cells are found.
    if any(
        'pyodide' in cb.attributes.get('classes', [])
        for cb in doctree.traverse(nodes.literal_block)
    ):
        return

    # Remove JS files
    pyodide_scripts = (
        app.config.nbsite_pyodide_conf['scripts'] +
        ['_static/run_cell.js', '_static/WorkerHandler.js']
    )

    context["script_files"] = [
        ii for ii in context["script_files"]
        if ii not in pyodide_scripts
    ]

    # Remove pyodide CSS files
    context["css_files"] = [
        ii for ii in context["css_files"]
        if ii not in ['_static/runbutton.css']
    ]


def setup(app):
    """Setup sphinx-gallery sphinx extension"""
    app.add_config_value('nbsite_pyodide_conf', DEFAULT_PYODIDE_CONF, 'html')

    app.connect('builder-inited', init_conf)
    app.connect('build-finished', write_worker)
    app.connect('html-page-context', html_page_context)
    app.connect('build-finished', PyodideDirective.terminate)

    app.add_directive('pyodide', PyodideDirective)

    app.add_css_file('runbutton.css')
    app.add_js_file('run_cell.js')
    app.add_js_file('WorkerHandler.js')

    return {
        'version': '0.5',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



import glob
import os
import re
import subprocess
import sys
from collections import ChainMap
from os.path import dirname

from .util import copy_files

DEFAULT_SITE_ORDERING = [
    "Introduction",
    "Getting Started",
    "User Guide",
    "Topics",
    "FAQ",
    "API"
]

def init(project_root='', doc='doc', theme=''):
    """
    Start an nbsite project: create a doc folder containing nbsite
    template files

    Use the theme argument to specify a particular theme to use as template
    """
    paths = _prepare_paths(project_root, doc=doc)
    if theme:
        copy_files(os.path.join(dirname(__file__), 'templates', theme),
                   paths['doc'])
    else:
        copy_files(os.path.join(dirname(__file__), 'templates', 'basic'),
                   paths['doc'])

hosts = {
    # no trailing slash
    'GitHub': 'https://raw.githubusercontent.com'
}

# TODO: clean up these fns + related arg parsing: parameterize, and
# maybe add task dependencies

def fix_links(output, inspect_links):
    args = ["nbsite_fix_links.py", output]
    if inspect_links:
        args.append( "--inspect-links")
    subprocess.check_call(args)


def build(what='html',
          output='builtdocs',
          project_root='',
          doc='doc',
          project_name='',
          host='GitHub',
          repo='',
          branch='main',
          org='',
          binder='none',
          examples='examples',
          examples_assets='assets',
          clean_dry_run=False,
          inspect_links=False,
          overwrite=False):
    """
    Build the site from the rst files and the notebooks

    Usually this is run after `nbsite scaffold`
    """
    env={'PROJECT_NAME':project_name,
         'PROJECT_ROOT':project_root if project_root!='' else os.getcwd(),
         'HOST':host,
         'REPO':repo,
         'BRANCH':branch,
         'ORG':org,
         'EXAMPLES':examples,
         'DOC':doc,
         'EXAMPLES_ASSETS':examples_assets,
         'BINDER':binder
         }
    merged_env = dict(os.environ, **env)
    none_vals = {k:v for k,v in merged_env.items() if v is None}
    if none_vals:
        raise Exception("Missing value for %s" % list(none_vals.keys()))

    paths = _prepare_paths(project_root, examples=examples, doc=doc, examples_assets=examples_assets)
    if overwrite:
        for path in glob.glob(os.path.join(paths['doc'], '**', '*.ipynb'), recursive=True):
            print('Removing evaluated notebook from {}'.format(path))
            os.remove(path)
    subprocess.check_call(["sphinx-build","-b",what,paths['doc'],output], env=merged_env)
    print('Copying json blobs (used for holomaps) from {} to {}'.format(paths['doc'], output))
    copy_files(paths['doc'], output, '**/*.json')
    copy_files(paths['doc'], output, 'json_*')
    if 'examples_assets' in paths:
        build_assets = os.path.join(output, examples_assets)
        print("Copying examples assets from %s to %s"%(paths['examples_assets'],build_assets))
        copy_files(paths['examples_assets'], build_assets)
    fix_links(output, inspect_links)
    # create a .nojekyll file in output for github compatibility
    subprocess.check_call(["touch", os.path.join(output, '.nojekyll')])
    if not clean_dry_run:
        print("Call `nbsite build` with `--clean-dry-run` to not actually delete files.")
    clean(output, clean_dry_run)

def clean(output, dry_run=False):
    if dry_run:
        subprocess.check_call(["nbsite_cleandisthtml.py",output])
    else:
        subprocess.check_call(["nbsite_cleandisthtml.py",output,'take_a_chance'])

def _prepare_paths(root,examples='',doc='',examples_assets=''):
    if root=='':
        root = os.getcwd()
    paths = {'project': os.path.abspath(root)}
    if examples!='':
        paths['examples'] = examples if os.path.isabs(examples) else os.path.join(paths['project'],examples)
    if doc!='':
        paths['doc'] = doc if os.path.isabs(doc) else os.path.join(paths['project'],doc)
    if examples_assets!='':
        paths['examples_assets'] = os.path.join(paths['examples'], examples_assets)
    return paths

def _is_root(x,root):
    return os.path.abspath(x) == os.path.abspath(root)

# TODO: needs cleaning up; currently just prototyping new behavior.
# TODO: rename to something like scaffold?
def generate_rst(
        project_name='',        # if not supplied, will default to repo
        project_root='',        # if not supplied, will default to os.getcwd()
        examples="examples",    # relative to project_root
        doc="doc",              # relative to project_root
#        title_formatter=None,
        host='GitHub',
        org='',                 # if not supplied, will default to project_name, or repo
        repo='',                # if not supplied, will default to project_name
        branch='main',
        offset=0,
        overwrite=False,
        nblink='bottom',
        skip='',
        keep_numbers=False,
        binder='none',
        disable_interactivity_warning=False
):
    """Auto-generates notebook-including rsts from notebooks in examples.

    titles
    ======

    rst file takes title from notebook filename in general, replacing
    underscores with spaces and converting to title case. E.g.

      * The_Title.ipynb -> The Title
      * 01_The_Title.ipynb -> The Title
      * 01_Hyphen-Conscious_Title.ipynb -> Hyphen-Conscious Title
      * 1_some_title.ipynb -> Some Title
      * 1_stripped_title.ipynb -> 1 Stripped Title  # with ``keep_numbers`` flag

    But: index.ipynb gets title from its containing directory in
    general, except at the root, where index.ipynb gets the
    project_name as title.


    toctree generation
    ==================

    If there's no index.rst already, the generated index.rst will
    contain a generated toctree, with entries sorted as follows:

      * number-prefixed items grouped before others, sorted first by
        number

      * after dealing with numeric prefixes, items in
        DEFAULT_SITE_ORDERING ("Introduction","Getting Started","User
        Guide","Topics","FAQ","API") grouped next, sorted in that
        order

      * remaining items appear at the end, sorted by text


    overriding titles and toctrees
    ==============================

    Generated rst files won't be overwritten on a subsequent run, so
    can be edited and committed.


    Other options
    =============

      * offset: allows to skip leading n cells
      * overwrite: will overwrite existing rst files [DANGEROUS]
      * keep_numbers: keeps leading numbers in title and path, rather than
        stripping them off.
      * ...

    """
    if repo == '' and project_name == '':
        raise ValueError("Must set at least one of repo or project-name")
    elif repo == '' or project_name == '':
        project_name = repo = project_name or repo

    if org == '':
        org = repo

    paths=_prepare_paths(project_root,examples=examples,doc=doc)

    print("Project='%s': Converting notebooks at '%s' to rst at '%s'..."%(project_name,paths['examples'],paths['doc']))
    for filename in glob.iglob(os.path.abspath(os.path.join(paths['examples'],"**","*.ipynb")), recursive=True):
        relpath = os.path.relpath(filename, paths['examples'])
        # TODO: decide what to do about gallery later
        if relpath.startswith('gallery'):
            continue
        if _should_skip(skip, relpath):
            print('...deliberately skipping', relpath)
            continue
        rst = os.path.splitext(os.path.join(paths['doc'], relpath))[0] + ".rst"
        rst = _path_and_order(rst, keep_numbers)[0]
        pretitle = _file2pretitle(rst)
        os.makedirs(dirname(rst), exist_ok=True)

        if pretitle != 'index':
            title = _to_title(pretitle)
        else:
            if _is_root(dirname(rst),paths['doc']):
                title = project_name
            else:
                title = _filepath2pretitle(rst,paths['doc'])
            title = _to_title(title, apply_title_case=True)

        if os.path.exists(rst):
            if not overwrite:
                print("...skipping %s"%rst)
                continue

        print("...writing %s"%rst)
        with open(rst, 'w') as rst_file:
            import nbsite
            rst_file.write('..\n   Originally generated by nbsite (%s):\n     %s\n   Will not subsequently be overwritten by nbsite, so can be edited.\n\n'%(nbsite.__version__,' '.join(sys.argv)))

            if binder in ['top', 'both']:
                add_binder(rst_file, org, repo, branch, relpath, examples)

            rst_file.write('*'*len(title)+'\n')
            rst_file.write(title+'\n')
            rst_file.write('*'*len(title)+'\n\n')

            if nblink in ['top', 'both']:
                add_nblink(rst_file, host, org, repo, branch, examples, relpath)
                rst_file.write('\n\n-------\n\n')

            rst_file.write(".. notebook:: %s %s" % (project_name,os.path.relpath(paths['examples'],start=dirname(rst))+'/'+relpath+"\n"))
            rst_file.write("    :offset: %s\n" % offset)
            if disable_interactivity_warning:
                rst_file.write("    :disable_interactivity_warning:\n")

            if pretitle=='index':
                rst_file.write("%s\n"%_toctree(dirname(filename), paths['examples'], keep_numbers))
            if nblink in ['bottom', 'both']:
                rst_file.write('\n\n-------\n\n')
                add_nblink(rst_file, host, org, repo, branch, examples, relpath)
            if binder in ['bottom', 'both']:
                add_binder(rst_file, org, repo, branch, relpath, examples)


def add_nblink(rst_file, host, org, repo, branch, examples, relpath):
    if all([host, org, repo, branch]):
        info = (hosts[host], org, repo, branch, examples,relpath)
        rst_file.write('`Right click to download this notebook from ' + host + '.'
                       ' <%s/%s/%s/%s/%s/%s>`_\n' % info)

def add_binder(rst_file, org, repo, branch, relpath, examples):
    rst_file.write("""
.. note:: Try live in your browser!

  |Binder| to run this notebook in your browser (no setup required).


.. |Binder| image:: https://mybinder.org/badge_logo.svg
   :target: https://mybinder.org/v2/gh/{org}/{repo}/{branch}?filepath={examples}/{relpath}

""".format(org=org,repo=repo,branch=branch,relpath=relpath,examples=examples))

def _should_skip(skip, filename):
    if skip == '':
        return False
    skip_patterns = [p.strip() for p in skip.strip('[]').split(',')]
    return any([re.match(p,filename,re.IGNORECASE) for p in skip_patterns])

def _to_title(name,apply_title_case=False):
    # allow to apply title case for directories, which pyviz has as lower case
    title = name.replace("_"," ")
    if apply_title_case:
        title = title.title()
    return title

def _filepath2pretitle(filepath,root):
    return os.path.relpath(dirname(filepath),start=root)

def _file2pretitle(file_):
    return os.path.splitext(os.path.basename(file_))[0]

def _title_key(title_tup):
    title, order = title_tup[0], title_tup[1]['order']
    if order is not None:
        return (0, order, title)
    elif title in DEFAULT_SITE_ORDERING:
        return (1, DEFAULT_SITE_ORDERING.index(title), title)
    else:
        return (1, float("inf"), title)

def _path_and_order(filepath, keep_numbers):
    num_name = os.path.basename(filepath)
    leading_num = re.match(r"^\d+", num_name)
    if not keep_numbers:
        name = re.split(r"^\d+( |-|_)", num_name)[-1]
        filepath = filepath.replace(num_name, name)
    return filepath, int(leading_num.group(0)) if leading_num else None

def _toctree(nbpath, examples_path, keep_numbers):
    tocmap = {'ipynb': {},'rst': {}}
    for ftype in ('ipynb','rst'):
        for f in glob.iglob(os.path.join(nbpath,'*.'+ftype), recursive=False):
            f, order = _path_and_order(f, keep_numbers)
            k = _file2pretitle(f)
            tocmap[ftype][_to_title(k)] = {'path': "<%s>"%k, 'order': order}

    tocmap['dir'] = {
        _to_title(k, apply_title_case=True): {'path': "<%s/index>"%k, 'order': None} for k in [
            _filepath2pretitle(x, examples_path) for x in glob.glob(os.path.join(nbpath, "**", "index.ipynb"))]}

    # index shouldn't explicitly appear in toctree...
    if 'index' in tocmap['rst'] or 'index' in tocmap['ipynb']:
        tocmap['rst'].pop('index', None)
        tocmap['ipynb'].pop('index', None)
    # ...except at root (where it gets called Introduction)
    if _is_root(nbpath,examples_path):
        if any(['Introduction' in tocmap[x] for x in tocmap]):
            raise ValueError("index will be shown as Introduction, but Introduction "
                             "already exists in %s" % examples_path)

        tocmap['rst']['Introduction'] = {'path': '<self>', 'order': -1}

    titles = ChainMap(*[tocmap[x] for x in tocmap])

    toctree = """
.. toctree::
    :titlesonly:
    :maxdepth: 2
"""
    for title, title_dict in sorted(titles.items(), key=_title_key):
        toctree += """
    %s %s"""%(title, title_dict['path'])

    return toctree
